from typing import Any
import pytest

from dagster_db import build_custom_duckdb_io_manager


@pytest.fixture(scope="module")
def resources() -> dict[str, Any]:
    return {
        "io_manager": build_custom_duckdb_io_manager().configured(
            {"database": "./.tests/database.duckdb"}
        ),
    }



import datetime as dt
from jinja2 import Template, StrictUndefined, Environment
from pandas import Timestamp
import typing as t


class SqlQuery:
    """
    Stores bindings for, and renders Jinja2-templated queries for use with SQL
    databases.
    Has type-specific rules for rendering bindings of different types.

    e.g. lists are converted into SQL syntax, SQL queries can be rendered
    inside other queries,
    and string values are wrapped in quotes so that they are not forgotten
    every time when working with strings.
    As a result, SqlExpr and SqlColumn have been created to allow no qoutes,
    or back-ticks (or another column name identifier) to be added when rendered.

    :param template_string: Jinja2-template string to be rendered.
    :**kwargs: bindings for the template.
    """

    def __init__(self, template_string: str, **kwargs):
        env = Environment(undefined=StrictUndefined)
        self.template: Template = env.from_string(template_string)
        self.bindings = kwargs

    def add_bindings(self, *args: t.Any, **kwargs: t.Any):
        self.bindings = self.bindings | dict(*args, **kwargs)

    def _render_value(self, value: t.Any) -> str:
        if isinstance(value, SqlQuery):
            return f"({value.render()})"
        elif isinstance(value, list):
            values_list = [self._render_value(v) for v in value]
            return f"({','.join(values_list)})"
        elif isinstance(value, (dt.datetime, dt.date, Timestamp)):
            return f"'{str(value)}'"
        elif isinstance(value, str):
            try:
                return self._render_value(Timestamp(value))
            except ValueError:
                return f"'{value}'"
        elif isinstance(value, (SqlExpr, SqlColumn)):
            return value.value
        else:
            return str(value)

    def render(self, *args: t.Any, **kwargs: t.Any) -> str:
        """Render all bindings in the query string."""
        bindings_original = self.bindings | dict(*args, **kwargs)

        bindings_curated = {}
        for key, value in bindings_original.items():
            bindings_curated[key] = self._render_value(value)

        return self.template.render(**bindings_curated)

    @property
    def markdown(self) -> str:
        return f"```sql\n{self.render()}\n```"


class SqlExpr:
    """
    Create a SQL expression that will be rendered as-is.
    """

    def __init__(self, value: str):
        self.value = value


class SqlColumn:
    """
    Refer to a column within SQL by using the databases column identifier
    (usually a back-tick).
    This will mean that quotes are not added when rendered.
    """

    def __init__(self, value: str, identifier: str = "`"):
        self.value = f"{identifier}{value}{identifier}"



from typing import Optional, Sequence, Type
import dagster as dg
from dagster_duckdb.io_manager import DuckDBIOManager, DuckDbClient

from dagster_db.io_managers.custom_db_io_manager import build_custom_db_io_manager
from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler
from dagster_db.type_handlers.duckdb_pandas import DuckDbPandasTypeHandler
from dagster_db.type_handlers.duckdb_polars import DuckDbPolarsTypeHandler
from dagster_db.type_handlers.duckdb_sql_query import DuckDbSqlQueryTypeHandler


def build_custom_duckdb_io_manager(
    type_handlers: Sequence[CustomDbTypeHandler] = [
        DuckDbPandasTypeHandler(),
        DuckDbPolarsTypeHandler(),
        DuckDbSqlQueryTypeHandler(),
    ],
    default_load_type: Optional[Type] = None,
    io_manager_name: str = "DuckDbIoManager",
) -> dg.IOManagerDefinition:
    """
    Create a configurable IO manager out of the `custom_db_io_manager` and
    type handlers.

    Configure with the `.configure()` method. `database` is a required key,
    `schema` is optional and can be set at the asset level.
    """
    return build_custom_db_io_manager(
        io_manager_base=DuckDBIOManager,
        db_client=DuckDbClient(),
        type_handlers=type_handlers,
        default_load_type=default_load_type,
        io_manager_name=io_manager_name,
    )



from typing import Any
import dagster as dg

from dagster_db import SqlQuery
from tests.helpers import result_metadata


def test_duckdb_sql_query_simple(resources: dict[str, Any]):
    @dg.asset
    def test_duckdb_sql_query_asset(context: dg.AssetExecutionContext) -> SqlQuery:
        query = SqlQuery("SELECT 1 AS a, 2 AS b")
        context.log.info(query.render())
        return query

    result = dg.materialize([test_duckdb_sql_query_asset], resources=resources)
    assert result.success

    metadata = result_metadata(result, "test_duckdb_sql_query_asset")
    assert "sample_obj" in metadata.keys()
    assert "sample_obj_db" in metadata.keys()
    assert "rows" in metadata.keys()
    assert "query_rendered" in metadata.keys()


def test_duckdb_sql_query_downstream(resources: dict[str, Any]):
    @dg.asset
    def test_duckdb_sql_query_asset(context: dg.AssetExecutionContext) -> SqlQuery:
        query = SqlQuery("SELECT 1 AS a, 2 AS b")
        return query

    @dg.asset
    def test_duckdb_sql_query_asset_downstream(
        context: dg.AssetExecutionContext,
        test_duckdb_sql_query_asset: SqlQuery,
    ) -> SqlQuery:
        query = SqlQuery(
            "SELECT *, 3 AS c FROM {{ test_duckdb_sql_query_asset }}",
            test_duckdb_sql_query_asset=test_duckdb_sql_query_asset,
        )
        return query

    result = dg.materialize(
        [test_duckdb_sql_query_asset, test_duckdb_sql_query_asset_downstream],
        resources=resources,
    )
    assert result.success



from typing import Any
import dagster as dg
import polars as pl

from tests.helpers import result_metadata


def test_duckdb_polars_simple(resources: dict[str, Any]):
    @dg.asset
    def test_duckdb_polars_asset(context: dg.AssetExecutionContext) -> pl.DataFrame:
        df = pl.DataFrame({"a": [1, 2, 3], "b": [2, 3, 4]})
        return df

    result = dg.materialize([test_duckdb_polars_asset], resources=resources)
    assert result.success

    metadata = result_metadata(result, "test_duckdb_polars_asset")
    assert "sample_obj" in metadata.keys()
    assert "sample_obj_db" in metadata.keys()
    assert "rows" in metadata.keys()


def test_duckdb_polars_downstream(resources: dict[str, Any]):
    @dg.asset
    def test_duckdb_polars_asset(context: dg.AssetExecutionContext) -> pl.DataFrame:
        df = pl.DataFrame({"a": [1, 2, 3], "b": [2, 3, 4]})
        return df

    @dg.asset
    def test_duckdb_polars_asset_downstream(
        context: dg.AssetExecutionContext,
        test_duckdb_polars_asset: pl.DataFrame,
    ) -> pl.DataFrame:
        test_duckdb_polars_asset_downstream = test_duckdb_polars_asset.with_columns(
            pl.lit(None).alias("c")
        )
        return test_duckdb_polars_asset_downstream

    result = dg.materialize(
        [test_duckdb_polars_asset, test_duckdb_polars_asset_downstream],
        resources=resources,
    )
    assert result.success



import datetime as dt

from jinja2.exceptions import UndefinedError
import pytest
from pandas import Timestamp
from dagster_db import SqlQuery, SqlExpr, SqlColumn


def test_sql_query_noop():
    base_query = "SELECT * FROM my_table"
    query = SqlQuery(base_query)
    assert query.render() == base_query


def test_sql_query_unbound():
    query = SqlQuery("SELECT {{ my_int }}")
    with pytest.raises(UndefinedError):
        query.render()


def test_sql_query_simple():
    query_int = SqlQuery("SELECT {{ my_int }}", my_int=1)
    assert query_int.render() == "SELECT 1"

    query_str = SqlQuery("SELECT {{ my_str }}", my_str="1")
    assert query_str.render() == "SELECT '1'"

    query_dt = SqlQuery("SELECT {{ my_dt }}", my_dt="2023-01-01")
    assert query_dt.render() == "SELECT '2023-01-01 00:00:00'"

    query_dt1 = SqlQuery("SELECT {{ my_dt }}", my_dt=dt.datetime(2023, 1, 1))
    assert query_dt1.render() == "SELECT '2023-01-01 00:00:00'"

    query_dt2 = SqlQuery("SELECT {{ my_dt }}", my_dt=dt.date(2023, 1, 1))
    assert query_dt2.render() == "SELECT '2023-01-01'"

    query_dt3 = SqlQuery("SELECT {{ my_dt }}", my_dt=Timestamp("2023-01-01"))
    assert query_dt3.render() == "SELECT '2023-01-01 00:00:00'"

    query_list = SqlQuery("SELECT {{ my_list }}", my_list=[1, 2])
    assert query_list.render() == "SELECT (1,2)"

    query_expr = SqlQuery("SELECT {{ my_expr }}", my_expr=SqlExpr("RANDOM()"))
    assert query_expr.render() == "SELECT RANDOM()"

    query_col = SqlQuery("SELECT {{ my_col }}", my_col=SqlColumn("my_col"))
    assert query_col.render() == "SELECT `my_col`"

    query_nested = SqlQuery("SELECT {{ query_int }}", query_int=query_int)
    assert query_nested.render() == "SELECT (SELECT 1)"


def test_sql_query_methods():
    query = SqlQuery("SELECT {{ my_int }}", my_int=1)
    assert query.bindings == {"my_int": 1}

    query = SqlQuery("SELECT {{ my_int }}")
    query.add_bindings({"my_int": 1})
    assert query.bindings == {"my_int": 1}

    query = SqlQuery("SELECT {{ my_int }}")
    query.add_bindings(my_int=1)
    assert query.bindings == {"my_int": 1}

    # no error
    query = SqlQuery("SELECT {{ my_int }}")
    query.render(my_int=1)



from typing import (
    Optional,
    Sequence,
    Type,
    Any,
    cast,
)
from pathlib import Path
import time

from dagster._core.storage.db_io_manager import DbIOManager, DbClient
import dagster as dg
import dagster._check as check

from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler


class CustomDbIOManager(DbIOManager):
    def __init__(
        self,
        *,
        type_handlers: Sequence[CustomDbTypeHandler],
        db_client: DbClient,
        database: str,
        schema: Optional[str] = None,
        io_manager_name: Optional[str] = None,
        default_load_type: Optional[Type] = None,
    ):
        self._handlers_by_type: dict[Optional[type[Any]], CustomDbTypeHandler] = {}  # type: ignore
        self._io_manager_name = io_manager_name or self.__class__.__name__
        for type_handler in type_handlers:
            for handled_type in type_handler.supported_types:
                check.invariant(
                    handled_type not in self._handlers_by_type,
                    f"{self._io_manager_name} provided with two handlers for the same type. "
                    f"Type: '{handled_type}'. Handler classes: '{type(type_handler)}' and "
                    f"'{type(self._handlers_by_type.get(handled_type))}'.",
                )
                self._handlers_by_type[handled_type] = type_handler

        Path(database).parent.mkdir(parents=True, exist_ok=True)

        self._db_client = db_client
        self._database = database
        self._schema = schema
        if (
            default_load_type is None
            and len(type_handlers) == 1
            and len(type_handlers[0].supported_types) == 1
        ):
            self._default_load_type = type_handlers[0].supported_types[0]
        else:
            self._default_load_type = default_load_type

    def handle_output(self, context: dg.OutputContext, obj: object) -> None:
        t0 = time.perf_counter()
        obj_type = type(obj)
        self._check_supported_type(obj_type)
        table_slice = self._get_table_slice(context, context)

        handler = self._handlers_by_type[obj_type]
        with self._db_client.connect(context, table_slice) as conn:
            obj_db = handler.db_safe_transformations(context, obj, conn)
            handler.validate_obj_db(context, obj_db, conn)

            context.log.debug("all validation successful")
            super().handle_output(context, obj)
            t1 = time.perf_counter()
            context.add_output_metadata(
                {
                    **handler.metadata(context, obj, obj_db, conn),
                    "io_time": dg.FloatMetadataValue(t1 - t0),
                }
            )

    def load_input(self, context: dg.InputContext) -> object:
        t0 = time.perf_counter()

        obj_type = context.dagster_type.typing_type
        if obj_type is Any and self._default_load_type is not None:
            load_type = self._default_load_type
        else:
            load_type = obj_type

        self._check_supported_type(load_type)

        table_slice = self._get_table_slice(
            context, cast(dg.OutputContext, context.upstream_output)
        )

        with self._db_client.connect(context, table_slice) as conn:
            obj = self._handlers_by_type[load_type].load_input(
                context, table_slice, conn
            )  # type: ignore  # (pyright bug)
            t1 = time.perf_counter()

            context.add_input_metadata(
                {
                    **self._handlers_by_type[load_type].metadata(
                        cast(dg.OutputContext, context.upstream_output),
                        obj,
                        None,
                        connection=conn,
                    ),
                    "io_time": dg.FloatMetadataValue(t1 - t0),
                }
            )
        return obj


def build_custom_db_io_manager(
    io_manager_base: dg.ConfigurableIOManagerFactory,
    db_client: DbClient,
    type_handlers: Sequence[CustomDbTypeHandler],
    default_load_type: Optional[Type] = None,
    io_manager_name: str = "CustomDbIoManager",
) -> dg.IOManagerDefinition:
    @dg.io_manager(config_schema=io_manager_base.to_config_schema())
    def duckdb_io_manager(init_context):
        return CustomDbIOManager(
            type_handlers=type_handlers,
            db_client=db_client,
            io_manager_name=io_manager_name,
            database=init_context.resource_config["database"],
            schema=init_context.resource_config.get("schema"),
            default_load_type=default_load_type,
        )

    return duckdb_io_manager



from dagster_db.io_managers.custom_db_io_manager import build_custom_db_io_manager
from dagster_db.io_managers.duckdb_io_manager import build_custom_duckdb_io_manager
from dagster_db.query.sql_query import SqlExpr, SqlQuery, SqlColumn
from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler
from dagster_db.type_handlers.duckdb_pandas import DuckDbPandasTypeHandler
from dagster_db.type_handlers.duckdb_polars import DuckDbPolarsTypeHandler
from dagster_db.type_handlers.duckdb_sql_query import DuckDbSqlQueryTypeHandler



import pandas as pd
import polars as pl
from duckdb import DuckDBPyConnection
from dagster_db.query.sql_query import SqlQuery
from dagster import get_dagster_logger

from typing import Optional, Type, TypeVar

T = TypeVar("T", list, tuple, pd.DataFrame, pl.DataFrame, DuckDBPyConnection)


def execute_duckdb(
    query: SqlQuery,
    connection: DuckDBPyConnection,
    obj: Optional[pl.DataFrame | pd.DataFrame] = None,
    return_type: Type[T] = DuckDBPyConnection,
) -> T:
    log = get_dagster_logger()
    query_rendered = query.render()
    log.debug(f"Running query:\n{query_rendered}")
    result = connection.execute(query_rendered)
    if return_type is list:
        return_value = result.fetchall()
    elif return_type is tuple:
        return_value = result.fetchone()
    elif return_type is pd.DataFrame:
        return_value = result.df()
    elif return_type is pl.DataFrame:
        return_value = result.pl()
    elif return_type is DuckDBPyConnection:
        return_value = result
    else:
        raise TypeError(f"Unsupported {return_type=}")

    return return_value  # type: ignore



import dagster as dg


def result_metadata(result: dg.ExecuteInProcessResult, node: str):
    materialisation = result.asset_materializations_for_node(node)
    metadata = materialisation[0].metadata
    return metadata



import dagster as dg
from typing import Sequence, Type
from dagster._core.storage.db_io_manager import TableSlice
import polars as pl
from duckdb import BinderException, DuckDBPyConnection, IOException
from dagster_duckdb.io_manager import DuckDbClient
from dagster._utils.backoff import backoff

from dagster_db.helpers.duckdb import execute_duckdb
from dagster_db.helpers.generic_db import table_slice_to_schema_table
from dagster_db.helpers.polars import get_sample_md, get_table_schema
from dagster_db.query.sql_query import SqlExpr, SqlQuery
from dagster_db.type_handlers.custom_type_handler import CustomDbTypeHandler


class DuckDbPolarsTypeHandler(CustomDbTypeHandler[pl.DataFrame, DuckDBPyConnection]):
    """
    Base-class for duckdb-polars type handler that handles read/write, and adds
    metadata, but doesn't include object validation or db_safe_transformations.
    """

    @property
    def supported_types(self) -> Sequence[Type[object]]:
        return [pl.DataFrame]

    def validate_obj_db(
        self, context, obj_db: pl.DataFrame, connection: DuckDBPyConnection
    ):
        return

    def db_safe_transformations(
        self, context, obj: pl.DataFrame, connection: DuckDBPyConnection
    ) -> pl.DataFrame:
        return obj

    def metadata(
        self,
        context: dg.OutputContext,
        obj: pl.DataFrame,
        obj_db: pl.DataFrame,
        connection: DuckDBPyConnection,
    ):
        return {
            "sample_obj": dg.MarkdownMetadataValue(get_sample_md(obj)),
            **(
                {"sample_obj_db": dg.MarkdownMetadataValue(get_sample_md(obj_db))}
                if obj_db is not None
                else {}
            ),
            "rows": dg.IntMetadataValue(obj.height),
            **(
                {"table_schema": dg.TableSchemaMetadataValue(get_table_schema(obj_db))}
                if obj_db is not None
                else {}
            ),
        }

    def _load_into_db(
        self, table_schema, obj: pl.DataFrame, connection: DuckDBPyConnection
    ):
        obj = obj.to_arrow()
        ctas_query = SqlQuery(
            "CREATE TABLE IF NOT EXISTS {{ table_schema }} AS SELECT * FROM obj;",
            table_schema=SqlExpr(table_schema),
        )
        execute_duckdb(ctas_query, connection, obj=obj)
        if not connection.fetchall():
            insert_query = SqlQuery(
                "INSERT INTO {{ table_schema }} SELECT * FROM obj",
                table_schema=SqlExpr(table_schema),
            )
            execute_duckdb(insert_query, connection, obj=obj)

    def handle_output(
        self,
        context: dg.OutputContext,
        table_slice: TableSlice,
        obj: pl.DataFrame,
        connection: DuckDBPyConnection,
    ):
        table_schema = table_slice_to_schema_table(table_slice)
        create_schema_query = SqlQuery(
            "CREATE SCHEMA IF NOT EXISTS {{ schema }}",
            schema=SqlExpr(table_slice.schema),
        )
        execute_duckdb(create_schema_query, connection)

        try:
            backoff(
                self._load_into_db,
                retry_on=(IOException,),
                kwargs={
                    "table_schema": table_schema,
                    "obj": obj,
                    "connection": connection,
                },
                max_retries=1,
            )
        except BinderException:
            obj_existing = self.load_input(context, table_slice, connection)
            msg = f"""`obj` incompatible with existing table

            `obj`
            {obj.glimpse()}

            `obj_existing`
            {obj_existing.glimpse()}
            """
            context.log.error(msg)

        return

    def load_input(
        self,
        context: dg.InputContext | dg.OutputContext,
        table_slice: TableSlice,
        connection: DuckDBPyConnection,
    ) -> pl.DataFrame:
        if table_slice.partition_dimensions and len(context.asset_partition_keys) == 0:
            raise ValueError(
                f"{table_slice.partition_dimensions=} incompatible with {context.asset_partition_keys=}"
            )

        query = DuckDbClient.get_select_statement(table_slice)
        obj = execute_duckdb(SqlQuery(query), connection, return_type=pl.DataFrame)
        return obj
